# Kosh transformer for scikit learn models
from .core import KoshTransformer


class Splitter(KoshTransformer):
    """SKL-based class to split dataset into test, train and validation"""
    types = {"numpy": ["numpy", ]}

    def __init__(self, train_size=None,
                 test_size=None,
                 validation_size=None,
                 splitter=None,
                 random_state=None,
                 n_splits=1,
                 *args, **kargs):
        """Initialize Splitt Transformer

        At least 2 of train/test/validation size are required, must add to 100%

        :param train_size: size of the dataset to reserve for training (.9 = 90% default)
        :type train_size: float
        :param test_size: size of the dataset to reserve for testing (.1 = 10% default)
        :type test_size: float
        :param validation_size: size of the dataset to reserve for validating (.0 = 0% default)
        :type validation_size: float
        :param splitter: SKL splitter to use to split
                         Same one will be used to first select training set
                         and then split again the rest between test and validation
                         default: sklearn.model_selection.ShuffleSplit
        :type splitter: sklearn.model_selection Splitter class
        :param random_state: random state for reproducibility
                             Controls the randomness of the training and
                             testing indices produced.
                             Pass an int for reproducible output across
                             multiple function calls.
        :type random_state: int
        :param n_splits: total number of split iteration to generate (default 1)
        :type n_splits: int
        :param groups: split data according to a third-party provided group.
                       This group information can be used to encode
                       arbitrary domain specific stratifications of the samples
                       as integers.
                       For instance the groups could be the year of collection
                       of the samples and thus allow for cross-validation against
                       time-based splits.
        :type groups: list or None
        :return: initialzed Splitter transformer
        :rtype: Splitter
        """

        if splitter is None:
            from sklearn.model_selection import ShuffleSplit
            splitter = ShuffleSplit
        if train_size is None and test_size is None and validation_size is None:
            train_size = .9
            test_size = .1
            validation_size = 0.
        elif train_size is None and validation_size is None:
            train_size = 1. - test_size
            validation_size = 0.
        elif train_size is None and test_size is None:
            test_size = validation_size
            train_size = 1. - test_size - validation_size
        elif test_size is None and validation_size is None:
            validation_size = 0.
            test_size = 1. - train_size
        elif validation_size is None:
            validation_size = 1. - train_size - test_size
        elif test_size is None:
            test_size = 1. - train_size - validation_size
        elif train_size is None:
            train_size = 1. - test_size - validation_size

        self.groups = kargs.pop("groups", None)
        if not (0. <= train_size <= 1.):
            raise ValueError("train size must be between 0 and 1")
        if not (0. <= test_size <= 1.):
            raise ValueError("test size must be between 0 and 1")
        if not (0. <= validation_size <= 1.):
            raise ValueError("validation size must be between 0 and 1")
        if train_size + test_size + validation_size > 1.:
            raise ValueError("You ask for a {}/{}/{} split which is more than 100%".format(
                train_size, test_size, validation_size))
        self.splitter = splitter(train_size=train_size,
                                 test_size=test_size+validation_size,
                                 random_state=random_state,
                                 n_splits=n_splits,
                                 *args, **kargs)
        self.validation_size = validation_size
        super(Splitter, self).__init__(train_size=train_size,
                                       test_size=test_size,
                                       validation_size=validation_size,
                                       splitter=splitter,
                                       random_state=random_state,
                                       n_splits=1,
                                       groups=self.groups, *args, **kargs)
        kargs.pop("test_size", None)
        kargs.pop("train_size", None)
        kargs["n_splits"] = 1
        kargs["random_state"] = random_state
        if validation_size != 0:
            self.validation_splitter = splitter(
                test_size=validation_size/(validation_size+test_size), *args, **kargs)

    def transform(self, input, format):
        """Split input data between n_splits sets of training/test/validation
        :param input: array from previous loader or transformer
        :type input: ndarray
        :param format: output format
        :type format: str
        :return: n_splits sets of train/test/validation
        :rtype: n_split list of train, test, validation ndarrays
        """
        out = list(self.splitter.split(input, groups=self.groups))
        if self.validation_size > 0:
            for i, [_, tmp_test] in enumerate(out):
                out[i] = list(out[i])  # Need to convert to list...
                tmp_test, tmp_validation = list(self.validation_splitter.split(tmp_test))[0]
                out[i][1] = tmp_test
            out[i].append(tmp_validation)
        return out


class StandardScaler(KoshTransformer):
    types = {"numpy": ["numpy", ]}

    def __init__(self, *args, **kargs):
        """SKL-based scaler transformer"""

        from sklearn.preprocessing import StandardScaler

        self.scaler = StandardScaler(*args, **kargs)
        super(StandardScaler, self).__init__(*args, **kargs)

    def transform(self, input, format):
        """calls the `fit_transform` function of the scaler on the input data
        :param input: array from previous loader or transformer
        :type input: ndarray
        :param format: output format
        :type format: str
        :return: scaled input data
        :rtype: ndarray
        """

        return self.scaler.fit_transform(input)


class SKL(KoshTransformer):
    """base class for SKL-based Kosh classifier
    This transformer returns either:
    * an estimator or
    * a set of labels/numpy arrays for each class found by the estimator
      If you chose to return the arrays (format=numpy) you can control
      how much data is return for each class/label
    When initiating the transformer you can pass any argument necessary for the SKL classifier initialization
    """
    types = {"numpy": ["estimator", "numpy"]}

    def __init__(self, *args, **kargs):
        """initialize Kosh classifier
        :param n_samples: number/percent of samples to
                          send back for each class
        :type n_samples: float
        :param sampling_method: units of percent random, or random_percent
                                units to returnthe number of samples
                                unit: return the first "n_samples"
                                percent: return the first n_samples % of the class
                                random_unit: return 'n_samples' random point from each class
                                random_percent: return n_samples % of the class randomly
        :type sampling_method: float
        :param random_state: random state for reproducibility
        :type random_state: int
        :param skl_class: SKL classifier
        :type skl_class: sklearn classifier
        :return: estimator from classifier.fit(input) function or
                 labels, list of ndarray with samples in each class
                        possibly sub-sampled via n_sample/sampling_method
        """
        kw = {}
        for arg in ["n_samples", "sampling_method", "random_state"]:
            setattr(self, arg, kargs.pop(arg, None))
            kw[arg] = getattr(self, arg)
        skl_class = kargs.pop("skl_class")
        self.skl_class = skl_class(*args, **kargs)
        kw.update(kargs)
        super(SKL, self).__init__(*args, **kargs)

    def transform(self, input, format):
        """If format is `numpy` scales the input data
        If format is `estimator` returns the estimator
        Possibly pads the ends with a value
        :param input: array from previous loader or transformer
        :type input: ndarray
        :param format: output format
        :type format: str
        :return: input taken over transformer's axis and indices
        """
        import numpy
        estimator = self.skl_class.fit(input)
        if format is not None and "estimator" in format.lower():
            return estimator
        labels = estimator.labels_
        out = []
        sorted_labels = sorted(set(labels))
        numpy.random.seed = self.random_state
        for each in sorted_labels:
            class_member_mask = (labels == each)
            if self.n_samples is not None:
                if self.sampling_method in [None, "unit"]:
                    out.append(input[class_member_mask][:self.n_samples])
                elif self.sampling_method == "percent":
                    out.append(input[class_member_mask][:int(
                        self.n_samples / 100. * input.shape[0])])
                elif self.sampling_method == "random":
                    indices = numpy.random.rand_integers(
                        0, input.shape - 1, self.n_samples)
                    out.append(input[class_member_mask][indices])
                elif self.sampling_method == "percent_random":
                    indices = numpy.random.rand_integers(
                        0, input.shape - 1, int(self.n_samples / 100. * input.shape[0]))
                    out.append(input[class_member_mask][indices])
            else:
                out.append(input[class_member_mask])
        return sorted_labels, out


class DBSCAN(SKL):
    """SKL-based DBSCAN Kosh classifier
    This transformer returns either:
    * an estimator or
    * a set of labels/numpy arrays for each class found by the estimator
      If you chose to return the arrays (format=numpy) you can control
      how much data is return for each class/label
    When initiating the transformer you can pass any argument necessary for the SKL classifier initialization
    """
    types = {"numpy": ["estimator", "numpy"]}

    def __init__(self, *args, **kargs):
        from sklearn.cluster import DBSCAN
        kargs["skl_class"] = DBSCAN
        super().__init__(**kargs)


class KMeans(SKL):
    """SKL-based KMeans Kosh classifier
    This transformer returns either:
    * an estimator or
    * a set of labels/numpy arrays for each class found by the estimator
      If you chose to return the arrays (format=numpy) you can control
      how much data is return for each class/label
    When initiating the transformer you can pass any argument necessary for the SKL classifier initialization
    """
    types = {"numpy": ["estimator", "numpy"]}

    def __init__(self, *args, **kargs):
        from sklearn.cluster import KMeans
        kargs["skl_class"] = KMeans
        super(KMeans, self).__init__(**kargs)



from .core import KoshLoader


class PandasLoader(KoshLoader):
    """ Kosh loader to load data using Pandas"""

    types = {"pandas/csv": ["DataFrame", ],
             "pandas/excel": ["DataFrame", ],
             "pandas/pickle": ["DataFrame", ],
             "pandas/table": ["DataFrame", ],
             "pandas/fwf": ["DataFrame", ],
             "pandas/clipboard": ["DataFrame", ],
             "pandas/json": ["DataFrame", ],
             "pandas/html": ["DataFrame", ],
             "pandas/xml": ["DataFrame", ],
             "pandas/hdf": ["DataFrame", ],
             "pandas/feather": ["DataFrame", ],
             "pandas/parquet": ["DataFrame", ],
             "pandas/orc": ["DataFrame", ],
             "pandas/sas": ["DataFrame", ],
             "pandas/spss": ["DataFrame", ],
             "pandas/sql_table": ["DataFrame", ],
             "pandas/sql_query": ["DataFrame", ],
             "pandas/sql": ["DataFrame", ],
             "pandas/gbq": ["DataFrame", ],
             "pandas/stata": ["DataFrame", ]}

    def _load_dataframe(self):

        import pandas as pd

        kwargs = self.obj.loader_kwargs

        if self._mime_type == 'pandas/csv':
            df = pd.read_csv(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/excel':
            df = pd.read_excel(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/pickle':
            df = pd.read_pickle(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/table':
            df = pd.read_table(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/fwf':
            df = pd.read_fwf(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/clipboard':
            df = pd.read_clipboard(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/json':
            df = pd.read_json(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/html':
            df = pd.read_html(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/xml':
            df = pd.read_xml(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/hdf':
            df = pd.read_hdf(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/feather':
            df = pd.read_feather(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/parquet':
            df = pd.read_parquet(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/orc':
            df = pd.read_orc(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/sas':
            df = pd.read_sas(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/spss':
            df = pd.read_spss(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/sql_table':
            df = pd.read_sql_table(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/sql_query':
            df = pd.read_sql_query(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/sql':
            df = pd.read_sql(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/gbq':
            df = pd.read_gbq(self.obj.uri, **kwargs)
        elif self._mime_type == 'pandas/stata':
            df = pd.read_stata(self.obj.uri, **kwargs)

        return df

    def open(self):
        """open/load matching file using `pandas.read_*()` method

        :return: Pandas DataFrame
        """
        df = self._load_dataframe()
        return df

    def extract(self):
        """extract return a feature from the loaded object.

        :param feature: variable to read from file
        :type feature: str
        :return: Pandas DataFrame
        """
        features = self.feature
        if not isinstance(features, list):
            features = [self.feature, ]

        df = self._load_dataframe()
        data = df[features]
        return data

    def list_features(self):
        """list_features list features in file,

        :return: list of features available in file
        :rtype: list
        """
        df = self._load_dataframe()
        features = df.columns.values.tolist()
        return features

    def describe_feature(self, feature):
        """describes the features

        :return: dictionary describing the feature
        :rtype: dict
        """
        df = self._load_dataframe()
        info = df[feature].describe()
        return info



from .core import KoshLoader


def read_chunk(f, end='\n'):
    out = ""
    cont = True
    while cont:
        r = f.read(1).decode()
        if r == end:
            return out
        out += r


class PGMLoader(KoshLoader):
    types = {"pgm": ["numpy", ]}

    def __init__(self, obj, **kargs):
        """PGMLoader for Kosh to be able to read in pgm image files
        :param obj: Kosh obj reference
        """
        super(PGMLoader, self).__init__(obj, **kargs)

    def open(self, mode="rb"):
        """open the pgm reader
        :param mode: mode to open the file in, defaults to 'r'
        :type mode: str, optional
        :return: Image file
        :rtype: file
        """
        return open(self.uri, mode)

    def extract(self):
        """get a feature

        :return: numpy array
        :rtype: numpy.ndarray
        """
        import numpy
        with open(self.uri, "rb") as f:
            magic = f.read(2).decode()
            if magic == 'P2':  # ASCII encoded
                _ = f.read(2)
                _ = read_chunk(f)
                dims = read_chunk(f)
                w, h = [int(x) for x in dims.split()]
                max_value = int(read_chunk(f))
                n = w*h
                data = numpy.array([int(x) for x in f.read().decode().split()]).reshape(h, w)
            elif magic == 'P5':  # binary encoded
                _ = f.read(1)
                dims = read_chunk(f)
                w, h = [int(x) for x in dims.split()]
                max_value = int(read_chunk(f))
                n = w*h
                data = numpy.frombuffer(
                    f.read(), dtype='u1' if max_value < 256 else 'u2', count=n).reshape(h, w)
            else:
                raise ValueError("Cannot read pgm magic number {magic}".format(magic=magic))

        return data

    def list_features(self):
        """list_features lists features available

        :return: list of features you can retrievea ["iamge", ] in our case
        :rtype: list
        """

        return ["image", ]

    def describe_feature(self, feature):
        """describe_feature describe the feature as a dictionary

        :param feature: feature to describe
        :type feature: str
        :return: dictionary with attributes describing the feature: 'size', 'format', 'max_value'
        :rtype: dict
        """
        with open(self.uri, "rb") as f:
            magic = f.read(2).decode()
            if magic == 'P2':  # ASCII encoded
                _ = f.read(2)
                _ = read_chunk(f)
                dims = read_chunk(f)
                w, h = [int(x) for x in dims.split()]
                max_value = int(read_chunk(f))
            elif magic == 'P5':  # binary encoded
                _ = f.read(1)
                dims = read_chunk(f)
                w, h = [int(x) for x in dims.split()]
                max_value = int(read_chunk(f))
            else:
                raise ValueError("Cannot read pgm magic number {magic}".format(magic=magic))

        return {"size": (h, w), "format": "pgm ({magic})".format(magic=magic), "max_value": max_value}



from filelock import FileLock, Timeout
from collections.abc import Generator
import time
import functools
import os
import logging

LOGGER = logging.getLogger()
LOGGER.setLevel(int(os.environ.get('LOCK_STRATEGIES_LOG_LEVEL', 30)))   # 30 is logging.WARNING
pid_map = {}


def lock_function(func):
    """
    Decorator to wrap functions with a LockStrategy.

    The lock is acquired at the start of the method and released after the
    method execution or after the entire generator is consumed.

    This is only intended to be used on class methods

    :param func: The method to be wrapped with the lock.
    :type func: Callable
    :return: The wrapped method with applied locking strategy.
    :rtype: Callable
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        lock_strategy = kwargs.get('lock_strategy', None)
        if lock_strategy is None:
            lock_strategy = NoLocking()
        return lock_strategy.threadsafe_call(func)(*args, **kwargs)
    return wrapper


def lock_method(func):
    """
    Decorator to wrap class methods with a LockStrategy.

    This decorator ensures that the method is executed with a locking
    mechanism provided by the LockStrategy. For methods that return
    generators, the lock will remain active throughout the iteration
    over the generator to ensure thread safety.

    The lock is acquired at the start of the method and released after the
    method execution or after the entire generator is consumed.

    :param func: The method to be wrapped with the lock.
    :type func: Callable
    :return: The wrapped method with applied locking strategy.
    :rtype: Callable
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kargs):
        @functools.wraps(func)
        def kosh_method(*args, **kargs):
            return func(self, *args, **kargs)
        with self.lock_strategy:
            result = self.lock_strategy.threadsafe_call(kosh_method)(*args, **kargs)
        return result
    return wrapper


class LockStrategy:
    """
    Base class for implementing different locking strategies.

    This class provides a common interface for various locking mechanisms, ensuring
    that derived classes can define their own locking behavior. The `LockStrategy`
    class implements the context manager protocol, allowing it to be used in `with`
    statements to acquire and release locks automatically.

    Derived classes should implement the `lock` and `unlock` methods to provide
    specific locking behavior, such as thread-based locks, file-based locks, or
    distributed locks.

    Example usage:
        >>> lock_strategy = SomeLockStrategy()
        >>> with lock_strategy:
        >>>     # Critical section of code
        >>>     pass

    Methods:
        lock(): Acquire the lock.
        unlock(): Release the lock.
        __enter__(): Enter the context manager, calling lock().
        __exit__(exc_type, exc_value, traceback): Exit the context manager, calling unlock().

    :param lock_strategy: A LockStrategy object to incorporate locking mechanisms.
    :type lock_strategy: LockStrategy
    """

    def lock(self):
        """Acquire the lock."""
        pass

    def unlock(self):
        """Release the lock."""
        pass

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.unlock()

    def threadsafe_call(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kargs):
            return func(*args, **kargs)
        return wrapper

    def wrap_generator(self, generator_func):
        @functools.wraps(generator_func)
        def wrapper(*args, **kargs):
            generator = generator_func(*args, **kargs)
            return self.lock_generator(generator)
        return wrapper

    def lock_generator(self, generator):
        while True:
            with self:
                try:
                    next_item = next(generator)
                except StopIteration:
                    break
            yield next_item


class NoLocking(LockStrategy):
    """
    A no-op LockStrategy
    """
    def __init__(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class RFileLock(LockStrategy):
    """
    A Re-entrant Filelock.
    """
    def __init__(self, num_tries=10, patience=1, lock_path=None, timeout=180):
        self.num_tries = num_tries
        self.patience = patience
        if lock_path is None:
            lock_path = os.path.join(os.path.expanduser("~"), ".kosh.lock")
        self.timeout = timeout
        self._lock = FileLock(lock_path, timeout=timeout)
        self.pid = os.getpid()
        # Keeps track of number of nested functions
        if self.pid not in pid_map:
            pid_map[self.pid] = 0
        self.functions = []

    def lock(self):
        import random
        LOGGER.info(msg=f"    {self.pid = } entering lock with {self.timeout = } secs & count {pid_map[self.pid]}...")
        exceptions = []
        for i in range(self.num_tries):
            try:
                if pid_map[self.pid] == 0:
                    self._lock.acquire()
                # Entering nested function
                pid_map[self.pid] += 1
                return

            except Timeout as e:
                exceptions.append(e)
                if i < (self.num_tries - 1):
                    msg = f"    Lock {self.timeout = } timed out. Waiting {self.patience} seconds before next attempt."
                    LOGGER.warning(msg=msg)
                    time.sleep(self.patience + random.random())  # random in case parallel calls retry at same time
        for e in exceptions:
            LOGGER.exception(msg=repr(e))
        raise exceptions[0]

    def unlock(self):
        # Exiting nested function
        pid_map[self.pid] -= 1
        # Back to original function
        if pid_map[self.pid] <= 0:
            self._lock.release()
            self.functions = []

    def threadsafe_call(self, func):
        import random

        @functools.wraps(func)
        def wrapper(*args, **kargs):
            exceptions = []
            num_tries = self.num_tries
            self.functions.append(str(func))
            while num_tries > 0:
                try:
                    LOGGER.info(msg=f'Entering function: {func} with {*args,} & { {k: v for k, v in kargs.items()} }')
                    result = func(*args, **kargs)
                    if isinstance(result, Generator):
                        result = self.lock_generator(result)

                    LOGGER.info(msg=f'Exiting function: {func}')
                    return result

                except Exception as e:
                    if any(func_error == "RETRIES COMPLETE" for func_error in self.functions):
                        raise
                    import traceback
                    error_stack = traceback.extract_stack()
                    tb = e.__traceback__
                    num_tries -= 1
                    exceptions.append(e)
                    msg = f"\nError in parent function {error_stack[0]}.\n"
                    msg += f"Exception {e} in child function {func.__name__}. "
                    msg += f'Details: {func} with {*args,} & { {k: v for k, v in kargs.items()} } '
                    msg += f"Retrying in {self.patience} seconds. {num_tries} retries remaining..."
                    trace = "\n\t\t".join([str(x) for x in traceback.extract_tb(tb)])
                    msg += f"\nError Stack:\n\t\t{trace}"
                    LOGGER.warning(msg=msg)
                    print(msg)  # For users without a logger
                    time.sleep(self.patience + random.random())  # random in case parallel calls retry at same time
            self.functions.append("RETRIES COMPLETE")
            exception_reprs = '\n'.join([repr(e) for e in exceptions])
            msg = f"Function {func.__name__} failed after {self.num_tries} attempts. "
            msg += f"Exception Traceback(s): \n{exception_reprs}"
            LOGGER.exception(msg=msg)
            raise exceptions[0]
        return wrapper


class OnlyRetry(RFileLock):
    """
    Only retry, no locking.
    """
    def __init__(self, num_tries=10, patience=1):
        self.num_tries = num_tries
        self.patience = patience
        self.pid = os.getpid()
        # Keeps track of number of nested functions
        if self.pid not in pid_map:
            pid_map[self.pid] = 0
        self.functions = []

    def __enter__(self):
        # Entering nested function
        pid_map[self.pid] += 1
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        # Exiting nested function
        pid_map[self.pid] -= 1
        # Back to original function
        if pid_map[self.pid] <= 0:
            self.functions = []
        pass

    def lock(self):
        pass

    def unlock(self):
        pass



from .core import KoshOperator  # noqa
from .core import typed_operator_with_kwargs, numpy_operator, typed_operator  # noqa
from .koshClustering import KoshCluster, KoshHopkins, KoshClusterLossPlot  # noqa
from .koshMaths import KoshLNorm  # noqa


from filelock import FileLock
import time
import functools
import os
import sys
import warnings


warnings.warn(("This version of threadsafe will be depreciated in Kosh 3.2. "
               "Use lock_strategies instead: examples/Example_ThreadSafe.ipynb."))


def threadsafe_call(num_tries, patience):
    """
    A custom thread lock decorator made specifically for Kosh calls.

    Parameters:
    -----------
    num_tries: int
        The number of allowable attempts.
    patience: float
        The number of seconds to wait between tries.

    Returns:
    --------
    A threadsafe version of the decorated function.
    """
    def decorator(func):

        @functools.wraps(func)
        def threadsafe(*args, **kw_args):
            exceptions = []
            for i in range(num_tries):
                try:
                    lock_path = os.path.join(os.path.expanduser("~"), ".kosh.lock")
                    lock = FileLock(lock_path, timeout=3600)

                    with lock:
                        ret = func(*args, **kw_args)
                        return ret

                except ValueError as e:
                    exceptions.append(e)
                    if i < (num_tries - 1):
                        msg = f"\nWARNING: function {func} threw exception within "
                        msg += f"threadsafe decorator. Waiting {patience} seconds "
                        msg += "before next attempt.\n"
                        sys.stderr.write(msg)
                        time.sleep(patience)

            for e in exceptions:
                sys.stderr.write(f"\n{e}")
            raise ValueError(exceptions[0])

        return threadsafe
    return decorator


def patient_find(num_tries, patience):
    """
    Kosh "find" methods can sometimes return empty results when they
    shouldn't. It's really unclear why/how this happens. This decorator
    allows us to perform the search multiple times if the initial results
    are empty.

    Parameters:
    -----------
    num_tries: int
        The number of allowable attempts.
    patience: float
        The number of seconds to wait between tries.

    Returns:
    --------
    A version of the "find" function that will re-attempt
    the find if the initial results are empty.
    """
    def decorator(func):

        @functools.wraps(func)
        def patient_call(*args, **kw_args):
            for i in range(num_tries):
                find_res = list(func(*args, **kw_args))

                if len(find_res) > 0:
                    return iter(find_res)
                else:
                    time.sleep(patience)

            msg = f"\nWARNING: patient_find failed {num_tries} attempts. Returning empty results.\n"
            sys.stderr.write(msg)

            return iter(find_res)

        return patient_call
    return decorator



from .core import KoshLoader


class SidreMeshBlueprintFieldLoader(KoshLoader):
    types = {"sidre_mesh_blueprint_fields":
             ["numpy", "sidre/path"]}

    def extract(self, *args, **kargs):
        import numpy
        import conduit
        import conduit.relay
        if not isinstance(self.feature, list):
            features = [self.feature]
        else:
            features = self.feature

        ioh = conduit.relay.io.IOHandle()
        ioh.open(self.uri, "sidre_hdf5")
        # look for self.feature
        bp_idx = conduit.Node()
        ioh.read(bp_idx, "root/blueprint_index")
        bp_path = conduit.Node()
        ndoms = conduit.Node()
        ioh.read(bp_path, "root/blueprint_index")

        out = []
        for feature in features:
            # split feature name back to mesh + field
            sp = feature.split("/")
            mesh = sp[0]
            field = "/".join(sp[1:])
            # get number of domains for this mesh
            ioh.read(
                ndoms,
                "root/blueprint_index/{}/state/number_of_domains".format(mesh))
            ndoms = ndoms.value()
            # get the path to the selected field in the bulk data
            pth = "root/blueprint_index/{}/fields/{}/path".format(mesh, field)
            if self.format == "sidre/path":
                out.append([ioh, pth])
                continue
            ioh.read(bp_path, pth)
            bp_path = bp_path.value()
            res = None
            vals = conduit.Node()
            for i in range(ndoms):
                dom_path = "%d/" % i
                dom_path += bp_path + "/values"
                ioh.read(vals, dom_path)
                npy_array = vals.value()
                if res is not None:
                    res = numpy.concatenate([res, npy_array])
                else:
                    res = npy_array
            out.append(res)
        if len(features) == 1:
            return out[0]
        else:
            return out

    def list_features(self):
        import conduit
        import conduit.relay
        ioh = conduit.relay.io.IOHandle()
        ioh.open(self.uri, "sidre_hdf5")
        # get the blueprint index
        bp_idx = conduit.Node()
        ioh.read(bp_idx, "root/blueprint_index")
        # enumerate meshes, and fields on each mesh
        res = []
        for mesh in bp_idx.children():
            for field in mesh.node()["fields"].children():
                res.append("/".join([mesh.name(), field.name()]))
        return res



from .core import KoshTransformer  # noqa
from .core import numpy_transformer, typed_transformer, typed_transformer_with_format  # noqa
from .npy import KoshSimpleNpCache  # noqa
from .skl import StandardScaler, KMeans, DBSCAN, Splitter  # noqa
from .sidre import SidreFeatureMetrics  # noqa
from kosh import kosh_cache_dir  # noqa


from kosh import kosh_cache_dir
import kosh
import random
import pickle
import os
import collections
import hashlib


def possible_ends(graph, start_nodes, end_nodes):
    """Finds all network ends that can be reached by all start nodes
    :param graph: The full graph
    :type graph: networkx.DiGraph (OrderedDiGraph on older version of networkx)
    :param start_nodes: Node to start paths from
    :type start_nodes: list of nodes
    :param end_nodes: Node to end paths from
    :type end_nodes: list of nodes
    :returns: list of possible end nodes
    :rtype: list
    """
    import networkx as nx
    ok_ends = []  # Matching end nodes for each start
    for start in start_nodes:
        ok_ends_this_start = []
        for end in end_nodes:
            try:
                nx.shortest_path(graph, start, end)
                ok_ends_this_start.append(end)
            except Exception:  # Can't get to this
                pass
        ok_ends.append(ok_ends_this_start)

    # Ok for each start we know know the ok end nodes
    # Which ones are common to every start?
    out = ok_ends[0]
    for ends in ok_ends[1:]:
        # now let's remove all end nodes that are not in the other paths
        for possible_end in list(out):
            if possible_end not in out:
                out.pop(possible_end)
    # out now contains the possible end nodes
    return out


def populate(G, node, output_formats, next_nodes):
    """Populates networkx
    :param G: networkx Graph to populate
    :type G: nx.Graph
    :param node: transformer to be chained, needs to have dict "types" attribute
    :type node: object with types attributes as a dictionary
    :param output_formats: output_format of the first node
    :type output_formats: list
    :param next_nodes: next set of transformers to add to graph
    :type next_nodes: object with 'types' attribute as a dictionary
    :returns: Nothing but the graph passed is updated
    """
    for format in output_formats:
        if format in list(next_nodes[0].types):
            this_node = (format, next_nodes[0], G.seed)
            G.add_edge(node, this_node)
            if len(next_nodes) > 1:
                populate(
                    G, this_node, next_nodes[0].types[format], next_nodes[1:])
            else:
                for final_fmt in next_nodes[0].types[format]:
                    final_node = (final_fmt, None, G.seed)
                    G.add_edge(this_node, final_node)


def find_network_ends(G, start=True, end=True):
    """Given an Ordered networkx.DiGraph finds start or end nodes or both.
    :param G: Network of interest
    :type G: networkx.DiGraph (OrderedDiGraph on older version of networkx)
    :param start: Are we searching for start nodes?
    :type start: bool
    :param end: Are we searching for end nodes?
    :type end: bool
    :returns: start and end node lists or just start/end node list
    :rtype: (tuple of) list
    :raises ValueError: if neither start or end is True
    """
    starters = []
    ends = []
    if not start and not end:
        raise ValueError("You need to set at least one of start/end to True")
    for node in G.nodes():
        if start and len(list(G.predecessors(node))) == 0:
            starters.append(node)
        if end and len(list(G.successors(node))) == 0:
            # Py2 seems to be returning these in random order
            # let's try to order this in consistent fashion
            # with predecessor's types order
            preds = list(G.predecessors(node))
            inserted = False
            for index, enode in enumerate(list(ends)):
                pred = preds[-1]
                pred_end = list(G.predecessors(enode))[-1]
                if pred[:1] == pred_end[:1]:
                    end_types = pred[1].types[pred[0]]
                    enode_index = end_types.index(enode[0])
                    node_index = end_types.index(node[0])
                    if node_index < enode_index:
                        ends.insert(index, node)
                        inserted = True
                        break
            if not inserted:
                ends.append(node)
    if start and not end:
        return starters
    elif end and not start:
        return ends
    else:
        return starters, ends


def apply_weight(G, output_format=None, weight_same=2., weight_output=3.):
    """Given a graph, lower the weight to edges that end in required format
    :param output_format: Desired output format, used to lower weight if edge ends in that format
    :type output_format: str
    :param weight_same: Weight to use if both end of an edge are the same format
    :type weight_same: float
    :param weight_output: Weight to use if end of an edge is the desired output format
    :type weight_output: float
    :return: Notne but the input graph is modified
    :rtype: None
    """
    for (n1, n2) in G.edges():
        weight = 1.
        # Does this edge connect identical formats?
        if n1[0] == n2[0]:
            weight /= weight_same

        # Does this edge ends with desired output format?
        if n2[0] == output_format and output_format is not None:
            weight /= weight_output
        G[n1][n2]["weight"] = weight


def get_seed(G, node, end_seed=None):
    """Assigns a new random seed to a node unless it is an end_seed, in which case we assign the Graph's seed
    :param G: Parent graph
    :type G: networkx.Graph
    :param node: Node of interest (is it and end node?)
    :type node: a graph node
    :param end_seed: The seed to assign if the node is an end seed. If None is passed then used parent Graph's seed
    :type seed: int (or None)
    :return: new seed for the node
    :rtype: int
    """
    if end_seed is None:
        end_seed = G.seed
    if len(list(G.successors(node))) == 0:
        seed = end_seed
    else:
        seed = random.random()
    return seed


class KoshExecutionGraph(object):
    types = {}

    def __len__(self):
        return len(self._graph)

    def __init__(self, *inputs, **kw):
        import networkx as nx
        graphs = []
        # Get a new seed
        self.seed = random.random()
        # Set the variable to receive results per index
        self.index_results = {}
        # Create a new merged graph
        try:
            new_graph = nx.OrderedDiGraph()
        except AttributeError:
            new_graph = nx.DiGraph()
        new_graph.seed = random.random()
        for i, G in enumerate(inputs):
            if isinstance(G, KoshExecutionGraph):
                G = G.execution_graph()
            elif not hasattr(G, "seed"):
                G.seed = random.random()
            # remember all graph we sent in
            graphs.append(G)
            new_graph.update(G)

        for mime in self.types:
            if not isinstance(mime, (list, tuple)):
                mime_list = [mime, ]
            else:
                mime_list = mime
            # Ok all inputs can be extracted to this thing input type
            for i, G in enumerate(graphs):
                if i >= len(mime_list):
                    output_format = mime_list[-1]
                else:
                    output_format = mime_list[i]
                for node in G.nodes():
                    if node == (output_format, None, G.seed):
                        # Connect last node to us
                        new_node = (node[0], self, self.seed)
                        pred = G.predecessors(node)
                        for n in pred:
                            new_graph.add_edge(n, new_node)
                        new_graph.remove_node(node)
                        # connect new node to export types
                        for export_type in self.types[mime]:
                            if not isinstance(export_type, (list, tuple)):
                                export_type = [export_type, ]
                            for export in export_type:
                                new_graph.add_edge(
                                    new_node, (export, None))
        # At this point we need to make sure there is a way out of this
        if isinstance(self, (kosh.operators.KoshOperator, kosh.transformers.KoshTransformer, kosh.loaders.KoshLoader)):
            start_nodes, end_nodes = find_network_ends(new_graph, start=True, end=True)
            out_types = set()
            for input_type in self.types:
                for output_type in self.types[input_type]:
                    out_types.add(output_type)
            # now let's go from each start node and see if we can go the an end node of an output type
            got_thru = False
            for output_type in out_types:
                for end_node in end_nodes:
                    if end_node[0] == output_type:
                        # ok can we go to this?
                        all_go = True
                        for start_node in start_nodes:
                            try:
                                _ = nx.shortest_path(new_graph, start_node, end_node)
                            except Exception:
                                all_go = False
                                break
                        if all_go:
                            got_thru = True
                            break
                if got_thru:
                    break

            if not got_thru:
                raise RuntimeError("Could not find an output format")

        self._graph = new_graph
        self.start_nodes, self.end_nodes = find_network_ends(new_graph)
        self.possible_end_nodes = possible_ends(new_graph, self.start_nodes, self.end_nodes)
        self.paths = collections.OrderedDict()

    def execution_graph(self, seed=None, verbose=False,
                        png_template="LOADER_GRAPH_{}"):
        """makes a new graph with unique seed
        Helps networkx differentiate between identical loaders/transformers/operators
        :param seed: seed to use for new graph
        :type seed: int
        :param verbose: verbose generation, also generates a png with the grap representation
                        Mostly used for debug purposes
        :type verbose: bool
        :param png_template: template to use to generate graph png in verbose mode
                             "_IN"/"_OUT" will be appended and seed will be fed
        :type png_template: str
        :return a new ordered graph with new seed
        :rtype: networkx.DiGraph
        """
        import networkx as nx
        try:
            G = nx.OrderedDiGraph()
        except AttributeError:
            G = nx.DiGraph()
        if seed is None:
            seed = random.random()
        G.seed = seed
        if verbose:
            try:
                if "DISPLAY" not in os.environ or os.environ["DISPLAY"] == "":
                    import matplotlib
                    matplotlib.use("agg", force=True)
                import matplotlib.pyplot as plt
                nx.draw(self._graph)
                plt.show()
                png_name = png_template + "_IN.png"
                plt.savefig(png_name.format(seed))
                plt.clf()
            except ImportError:
                raise RuntimeError("Could not import matplotlib, will not plot anything")
        used_nodes = collections.OrderedDict()
        for (n1, n2) in self._graph.edges():
            if n1 in used_nodes:
                # we already generated a new random number for that node
                N1 = used_nodes[n1]
            else:
                # Never seen that node
                seed = get_seed(self._graph, n1, G.seed)
                N1 = n1[0], n1[1], seed
                used_nodes[n1] = N1
            if n2 in used_nodes:
                # we already generated a new random number for that node
                N2 = used_nodes[n2]
            else:
                # Never seen that node
                seed = get_seed(self._graph, n2, G.seed)
                N2 = n2[0], n2[1], seed
                used_nodes[n2] = N2
            G.add_edge(N1, N2)
        if verbose:
            nx.draw(G)
            plt.show()
            png_name = png_template + "_OUT.png"
            plt.savefig(png_name.format(seed))
            plt.clf()
        return G

    def __getitem__(self, key):
        """Very bare bone get item function
        It is highly recommended to re-implement this.
        Calls traverse then __getitem__ on the result of traverse
        :param key: key to access
        :type key: object (usually int, slice or str)
        """
        return self.traverse(__getitem_key__=key)

    def traverse(self, format=None, *args, **kargs):
        """Traverse the execution graph and returns data
        :param format: desired output format
        :type format: str
        """
        import networkx as nx
        G = self._graph
        start_nodes, _ = self.start_nodes, self.end_nodes

        possible_end_nodes = self.possible_end_nodes

        if len(possible_end_nodes) == 0:
            raise RuntimeError("This graph cannot be traversed to a single end node from each start. Aborting")
        # We first need to determine the output_format
        if format is None:  # User lets us pick
            format = possible_end_nodes[0][0]

        if format not in [end[0] for end in possible_end_nodes]:
            raise ValueError("Cannot output in format {}".format(format))

        # Which node is our exit node?
        for end_node in possible_end_nodes:
            if end_node[0] == format:
                break

        if format not in self.paths:
            # Ok now let's apply the weights
            apply_weight(G, output_format=format)

            # And get the shortest path(s)
            # For each entry path
            pths = []
            for start_node in start_nodes:
                pths.append(nx.shortest_path(G, start_node, end_node))
            self.paths[format] = pths
        else:
            pths = self.paths[format]

        # Ok let's generate the new network with only the paths
        try:
            out = nx.OrderedDiGraph()
        except AttributeError:
            out = nx.DiGraph()
        out.seed = G.seed
        for pth in pths:
            for i, node in enumerate(pth[:-1]):
                if pth[i+1][1] is not None:
                    pth[i+1][1].parent = node[1]
                out.add_edge(node, pth[i + 1])

        # We can now travel back the pth to obtain
        # the data.
        return self._operate(out, end_node, format, **kargs)

    __call__ = traverse

    def _operate(self, graph, node, output_format, **kargs):
        """Actual bells and whistles to get the data
        :param graph: The graph to follow in order to get the data
        :type graph: KoshExecutionGraph
        :param node: node to process (get inputs and call extract/transform/operate func)
        :type node: node in the network
        :param out_format: The desired output_format
        :type output_format: str
        :param kargs: key arguments that will be passed to loaders (start of each path)
        :type kargs: dict
        :returns: Data
        """
        getitem_key = kargs.pop("__getitem_key__", slice(None, None, None))
        cache_file_only = kargs.pop("cache_file_only", False)
        use_cache = kargs.pop("use_cache", False)
        cache_dir = kargs.pop("cache_dir", kosh_cache_dir)
        _, end = find_network_ends(graph, start=True, end=True)
        previous = list(graph.predecessors(node))
        if len(previous) == 0:
            # Ok we are at the start e.g a loader
            node[1].cache_file_only = cache_file_only
            node[1].use_cache = use_cache
            node[1].cache_dir = cache_dir
            node[1]._user_passed_parameters = (None, kargs)
            if not isinstance(getitem_key, slice) or getitem_key != slice(None, None, None):
                if "__getitem__" in node[1].__class__.__dict__:
                    out = node[1][getitem_key]
                else:
                    out = node[1].extract_(format=output_format)[getitem_key]
            else:
                out = node[1].extract_(format=output_format)
            return out
        else:
            inputs = ()
            for input_index, prev in enumerate(previous):
                kargs2 = kargs.copy()
                do_res = True
                if hasattr(node[1], "__getitem_propagate__") and getitem_key != slice(None, None, None):
                    new_keys = node[1].__getitem_propagate__(getitem_key, input_index=input_index)
                    kargs2["__getitem_key__"] = new_keys
                    if new_keys is None:
                        res = getattr(node[1], "index_results", {}).get(input_index, None)
                        do_res = False
                elif type(self) is kosh.exec_graphs.core.KoshExecutionGraph or node == end[0]:
                    new_keys = False
                    kargs2["__getitem_key__"] = getitem_key
                else:
                    new_keys = None
                    kargs2["__getitem_key__"] = slice(None, None, None)
                if do_res:
                    res = prev[1]._operate(graph, prev, node[0], **kargs2)
                    if new_keys is None and getitem_key != slice(None, None, None):
                        res = res[getitem_key]
                inputs += (res,)
            if node == end[0]:
                # this is the last one, do not call operate on it
                # we are done
                return inputs[0]
            if hasattr(self, "operate_"):
                out = self.operate_(*inputs, format=output_format)
                return out
            elif hasattr(self, "transform_"):
                out = self.transform_(*inputs, format=output_format)
                return out
            elif isinstance(self, kosh.exec_graphs.core.KoshExecutionGraph):
                return inputs
            else:
                raise RuntimeError(
                    "Did not know which function to send inputs to. Aborting")

    def update_signature(self, *args, **kargs):
        """Updated the signature based to a set of args and kargs
        :param *args: as many arguments as you want
        :type *args: list
        :param **kargs: key=value style argmunets
        :type **kargs: dict
        :return: updated signature
        :rtype: str
        """
        signature = hashlib.sha256(self.signature.encode())
        for arg in args:
            signature.update(repr(arg).encode())
        for kw in kargs:
            signature.update(repr(kw).encode())
            signature.update(repr(kargs[kw]).encode())
        return signature.hexdigest()

    def show_cache_file(self, input, format):
        """Given a set of input and format returns the unique signature used for cache file
        :param input: set of input passed from loader or previous transformer
        :type input: object
        :param format: desired output format
        :type format: str
        :return: The unique signature
        :rtype: str
        """
        signature = self.update_signature(input, format)
        return os.path.join(self.cache_dir, signature)

    def save(self, cache_file, *content):
        """Pickles some data to a cache file
        :param cache_file: name of cache file, will be joined with self.cache_dir
        :type cache_file: str
        :param content: content to save to cache
        :type content: object
        """
        with open(os.path.join(self.cache_dir, cache_file), "wb") as f:
            for sv in content:
                pickle.dump(sv, f)

    def load(self, cache_file):
        """Loads content from cache
        :param cache_file: name of cache fileA will be joined with self.cache_dir
        :type cache_file: str
        :return: unpickled data
        :rtpye: object
        """
        with open(os.path.join(self.cache_dir, cache_file), "rb") as f:
            cont = True
            data = []
            while cont:
                try:
                    data.append(pickle.load(f))
                except Exception:
                    cont = False
        if len(data) == 1:
            return data[0]
        else:
            return data

    def get_input_loaders(self):
        """Return a generator of the originating loaders for the inputs.
        It will crawl the graph backward when
        it encounters transformers or operators.
        :return: generator of Kosh datasets
        :rtype: KoshLoader generator
        """
        for start_node in self.start_nodes:
            loader = start_node[1]
            yield loader

    def get_input_datasets(self):
        """Return a generator of the originating datasets for the inputs.
        It will crawl the graph backward when
        it encounters transformers or operators.
        :return: generator of Kosh datasets
        :rtype: KoshDataset generator
        """
        for start_node in self.start_nodes:
            loader = start_node[1]
            yield loader.get_requestor()

    def describe_entries(self):
        """Return a generator of describe_feature for each entry feature.
        It will crawl the graph backward when
        it encounters transformers or operators.
        If a loader did not implement `describe_feature` an empty dictionary
        will be used instead.
        :return: generator of info dictionaries
        """
        for start_node in self.start_nodes:
            loader = start_node[1]
            try:
                info = loader.describe_feature(loader.feature)
            except NotImplementedError:
                info = {}
            yield info



# Ultra files loader contributed by Josh Kallman 5/6/2020
from .core import KoshLoader
import sys
import os
sys.path.append("/usr/gapps/pydv/current")  # noqa


class UltraLoader(KoshLoader):
    """Kosh Loader for ultra files"""
    types = {"ultra": ["dict", "numpy"]}

    def __init__(self, obj, **kargs):
        super(UltraLoader, self).__init__(obj, **kargs)
        self.curves = None

    def load_curves(self):
        # pydv import matplotlib.pyplot
        # on some systems with no X forwarding this causes
        # an uncatchable error.
        # Setting the matplotlib backend to a windowless
        # backend fixes this.
        if "DISPLAY" not in os.environ or os.environ["DISPLAY"] == "":
            import matplotlib
            matplotlib.use("agg", force=True)
        try:
            import pydvpy as pydvif
        except ImportError:
            import pydv
            sys.path.append(pydv.__path__[0])
            import pydv.pydvpy as pydvif
        self.curves = pydvif.read(self.uri)

    def load_from_ultra(self, variable):
        """Load variables from an ultra file
        :param variable: variables to load
        :type variable: list or str
        :return Dictionary containing 'x-axis' and 'y-axis' for each variable
        :rtype: dict
        """
        if self.curves is None:
            self.load_curves()
        if not isinstance(variable, (list, tuple)) and variable is not None:  # only one variable requested
            variable = [variable, ]

        variables = {}

        if variable is None:  # all curves
            for c in self.curves:
                name = c.name
                variables[name] = {}
                variables[name]['x-axis'] = c.x
                variables[name]['y-axis'] = c.y
        else:
            for var in variable:
                for c in self.curves:
                    name = c.name
                    if name == var:
                        variables[name] = {}
                        variables[name]['x-axis'] = c.x
                        variables[name]['y-axis'] = c.y
                        break

        return variables

    def extract(self, *args, **kargs):
        """Extract a feature"""
        return self.load_from_ultra(self.feature)

    def open(self):
        """open/load matching ultra file

        :return: Dictionary containing 'x-axis' and 'y-axis' for each variable
        """
        return self.load_from_ultra(None)

    def list_features(self):
        """List features available in ultra file"""
        variables = []
        if self.curves is None:
            self.load_curves()
        for curve in self.curves:
            variables.append(curve.name)
        return variables

    def describe_feature(self, feature):
        """Describe a feature

        :param feature: feature to describe
        :type feature: str
        :return: dictionary with attributes describing the feature:
                 'name', 'size', 'first_time', 'last_time', 'min', 'max', 'type'
        :rtype: dict
        """
        info = {"name": feature}
        if self.curves is None:
            self.load_curves()
        for c in self.curves:
            if c.name == feature:
                info["size"] = len(c.x)
                info["first_time"] = c.x[0]
                info["last_time"] = c.x[-1]
                info["min"] = min(c.y)
                info["max"] = max(c.y)
                info["type"] = c.y.dtype
                break
        return info



from .Clustering import Cluster, makeBatchClusterParallel  # noqa
from .Clustering import SerialClustering, ParallelClustering, SubsampleWithLoss  # noqa
from .Clustering import numpyParallelReader, scaleDataParallel # noqa
from .Clustering import GetMaxLoss  # noqa


try:
    import conduit
    has_conduit = True
except ImportError:
    has_conduit = False
from .utils import get_ids_for_rank, get_mpi_tools
from .core import KoshTransformer


class SidreFeatureMetrics(KoshTransformer):
    """Numpy-based Metrics
    this returns a dictionary of metrics for a feature (sent as input)
    """
    types = {"sidre/path": ["dict"]}

    def __init__(self, *args, **kargs):
        if not has_conduit:
            raise RuntimeError(
                "Could not import conduit, Conduit-based transformers are not available")
        super(SidreFeatureMetrics, self).__init__(*args, **kargs)

    def transform(self, input_, format):
        import numpy
        rank, size, comm = get_mpi_tools()
        from mpi4py import MPI
        if rank != 0:
            stats = None
        else:
            stats = {}
        ioh, pth = input_

        sp0 = pth.split("/fields")[0]

        mesh = sp0.split("/")[-1]
        # First let's figure out the number of domains
        ndoms = conduit.Node()
        ioh.read(
            ndoms,
            "root/blueprint_index/{}/state/number_of_domains".format(mesh))
        ndoms = int(ndoms.value())

        # which domains do I read?
        my_domains = get_ids_for_rank(ndoms)

        bp_path = conduit.Node()
        ioh.read(bp_path, pth)
        bp_path = bp_path.value()
        vals = conduit.Node()
        # now read the data for this rank
        sum = 0
        N = 0
        mn = 1.e999
        mx = -1.e999
        rank_vals = None
        dtype = -1
        for domain in my_domains:
            dom_path = "{}/".format(domain)
            dom_path += bp_path + "/values"
            ioh.read(vals, dom_path)
            data = vals.value()
            mn = min(data.min(), mn)
            mx = max(data.max(), mx)
            N += len(data.flat)
            sum += data.sum()
            if rank_vals is None:
                rank_vals = data
            else:
                rank_vals = numpy.concatenate((rank_vals, data))
                dtype = rank_vals.dtype

        if rank == 0:
            for i in range(1, size):
                N += comm.recv(source=i, tag=1)
                sum += comm.recv(source=i, tag=2)
                mn = min(mn, comm.recv(source=i, tag=3))
                mx = max(mx, comm.recv(source=i, tag=4))
                comm.send(dtype, dest=i, tag=5)
        else:
            comm.send(N, dest=0, tag=1)
            comm.send(sum, dest=0, tag=2)
            comm.send(mn, dest=0, tag=3)
            comm.send(mx, dest=0, tag=4)
            dtype = comm.recv(source=0, tag=5)

        # Histogram from: https://ascent.readthedocs.io/en/latest/Tutorial_CloverLeaf_Demos.html
        # compute bins on global extents
        bins = numpy.linspace(mn, mx)

        if rank_vals is not None:  # DAta on this rank
            # get histogram counts for local data
            hist, bin_edges = numpy.histogram(rank_vals, bins=bins)
        else:
            hist = numpy.zeros((len(bins) - 1), dtype=dtype)

        # declare var for reduce results
        hist_all = numpy.zeros_like(hist)
        # sum histogram counts with MPI to get final histogram
        comm.Allreduce(hist, hist_all, op=MPI.SUM)

        if rank == 0:
            stats["histogram"] = hist
            stats["min"] = float(mn)
            stats["max"] = float(mx)
            stats["mean"] = float(sum) / float(N)

        # Now we can compute the std dev
        std = 0.
        if stats is not None:
            # We have data on this rank
            for domain in my_domains:
                dom_path = "{}/".format(domain)
                dom_path += bp_path + "/values"
                ioh.read(vals, dom_path)
                std += ((vals.value() - stats["mean"])**2).sum()

        if rank == 0:
            for i in range(1, size):
                std += comm.recv(source=i, tag=6)
            stats["std"] = float(numpy.sqrt(std / N))
        else:
            comm.send(std, dest=0, tag=6)

        stats = comm.bcast(stats, root=0)
        return stats



# Setup script for dkosh repo
from setuptools import setup, find_packages
from subprocess import Popen, PIPE
import shutil
import os


exec(open("./kosh/current_version.py").read())
version = current_version  # noqa
sha = None

git_describe_process = Popen(
    ("git",
     "describe",
     "--tags"),
    stdout=PIPE,
    stderr=PIPE)
if os.path.isdir(os.path.join(os.path.dirname(
    os.path.abspath(__file__)),'.git')):
    try:
        out, _ = git_describe_process.communicate()
        trial_version = out.decode("utf-8")
        sp = trial_version.split("-")
        trial_version = sp[0]
        # Clean tag?
        if len(sp) != 0:
            commits = sp[1]
            sha = sp[2]
            trial_version += "."+commits
        else:
            sha = None
        version = trial_version
    except Exception:
        pass

description="Manages Data Store with External Bulk Data"
if sha is not None:
    description += " (sha: {})".format(sha)

with open("README.md", "r") as fh:
    long_description = fh.read()
if os.path.exists("scripts/kosh"):
    os.remove("scripts/kosh")
shutil.copyfile("kosh/kosh_command.py", "scripts/kosh")
setup(name="kosh",
      version=version,
      description=description,
      url="https://github.com/LLNL/Kosh",
      author="Charles Doutriaux, Renee Olson",
      author_email="doutriaux1@llnl.gov",
      long_description=long_description,
      long_description_content_type="text/markdown",
      license="MIT",
      packages=find_packages(),
      scripts=["scripts/kosh",
               "kosh/kosh_command.py",
               "scripts/sbang",
               ],
      zip_safe=False,
      install_requires=[
          'llnl-sina >=1.15',
          'networkx>=2.6',
          'numpy>=1.20',
          'scipy',
          'h5py>=3',
          'scikit-learn>=1.0.2',
          'pandas',
          'hdbscan',
          'matplotlib',
          'tqdm',
          'filelock'
      ],
      classifiers=[
          "Programming Language :: Python",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      )




from __future__ import absolute_import
import collections
import os
import kosh
import hashlib
import random
import warnings
import pickle
from .wrapper import KoshScriptWrapper # noqa
from . import lock_strategies
try:
    import orjson
except ImportError:
    import json as orjson  # noqa


class KoshPickler(object):
    def __init__(self, encode_types=["windows-1252", "latin1"]):
        self.encode_types = encode_types

    def dumps(self, data):
        success = False
        for encode_type in self.encode_types:
            try:
                encoded = pickle.dumps(data).decode(encode_type)
                self.loads(encoded)
                success = True
                break
            except Exception:
                pass
        if success:
            return encoded
        else:
            raise RuntimeError("Could not pickle")

    def loads(self, data):
        success = False
        for encode_type in self.encode_types:
            try:
                loaded = pickle.loads(data.encode(encode_type))
                success = True
                break
            except Exception:
                pass
        if success:
            return loaded
        else:
            raise RuntimeError("Could not load pickled")


def find_curveset_and_curve_name(name, rec):
    """Given a curveset or curveset+curve name,
    returns all matching curve_sets and curves combinations
    Assumes curveset and curve are separated by a /
    curve_sets that exactly match the name return (name, None)
    :param name: Name to parse
    :type name: str
    :param rec: sina record where to look for curves
    :type rec: sina record
    :return: All possible combinations of (curveset,curve) that match name
    :rtype: tuple of tuples
    """
    sp = name.split("/")
    possibilities = ()
    for i in range(len(sp)+1):
        curve_set = "/".join(sp[:i])
        if curve_set in rec["curve_sets"]:
            cs = rec["curve_sets"][curve_set]
            curve_name = "/".join(sp[i:])
            if curve_name in cs["dependent"] or curve_name in cs["independent"]:
                possibilities += ((curve_set, curve_name), )
            if curve_set == name:  # ok we asked for the curve_set
                possibilities += ((curve_set, None),)
    return possibilities


def merge_datasets_handler(target_dataset, imported_dataset, section="data", **kargs):
    """When importing a dataset, checks if the imported dataset has
    attributes that match the one in the dataset already in this store.
    If attributes values conflict then we use 'handling_method to resolve the conflict

    The store_dataset is not updated here, we return a list of attributes/values pairs resolving the conflict

    :param target_dataset: The dataset that will received the merge
    :type target_dataset: kosh.KoshDataset
    :param imported_dataset: The dataset we are trying to merge into target_dataset or its attributes/values dictionary
    :type imported_dataset: kosh.KoshDataset or dict
    :param section: The section being updated (data, user_defined, curves, etc...)
    :type section: str
    :param handling_method: How do we handle conflicts?
                            None, "conservative": Error exit
                            "preserve": Keep value from target_dataset
                            "overwrite": Use value from imported dataset
    :returns: Dictionary of attribute/value that the target_dataset should have
    :rtype: dict
    """
    if section != "data":
        raise ValueError("This handler cannot handle non 'data' section")

    handling_method = kargs.pop("handling_method", None)

    target_dict = target_dataset.list_attributes(dictionary=True)

    if not isinstance(imported_dataset, dict):
        imported_dataset = imported_dataset.list_attributes(dictionary=True)

    # We cannot set _associatated_data_ anyway and if it comes last (py2) it
    # prevents updating the db
    imported_dataset.pop("_associated_data_", None)
    # Creator is a pain looks like it comes with id at times and actual user name at others
    # TODO: take a closer look in a separate MR
    if "creator" in target_dict:
        imported_dataset.pop("creator", None)

    for attribute, value in imported_dataset.items():
        if attribute in target_dict:
            if target_dict[attribute] != value:
                if handling_method in [None, "conservative"]:
                    msg = "Trying to import dataset with attribute '{}'".format(
                        attribute)
                    msg += " value : {}. ".format(value)
                    msg += "But value for this attribute in target is '{}'".format(
                        target_dict[attribute])
                    raise ValueError(msg)
                elif handling_method == "overwrite":
                    # Do we want a warning here?
                    # handling says use new value
                    target_dict[attribute] = value
                elif handling_method == "preserve":
                    # Do we want a warning here?
                    # We preserve so not changing the target value
                    pass
                else:
                    raise ValueError(
                        "Unknown 'handling_method': {}".format(handling_method))
        else:
            # New attribute let's add it
            target_dict[attribute] = value
    return target_dict


def gen_labels(G):
    """Generates labels to draw on networkx plots of a graph
    :param G: Network to generate labels from
    :type G: networkx.DiGraph (OrderedDiGraph on older version of networkx)
    :returns: labels for this graph
    :rtype: dict
    """
    labels = {}
    cont = True
    nodes = list(G.nodes())
    N = len(nodes)
    while cont:
        for node in nodes:
            if G.nodes[node].get("depth", None) is not None:
                continue
            pre = list(G.predecessors(node))
            suc = list(G.successors(node))
            if len(pre) == 0:
                G.nodes[node]["depth"] = 0
            else:
                for pnode in pre:
                    if G.nodes[pnode].get("depth", None) is not None:
                        G.nodes[node]["depth"] = G.nodes[pnode]["depth"] + 1
            if len(suc) == 0:
                G.nodes[node]["depth"] = -1
        total = 0
        for node in nodes:
            if G.nodes[node].get("depth", None) is not None:
                total += 1
        if total == N:
            cont = False
    for node in nodes:
        depth = G.nodes[node]["depth"]
        if depth == 0:
            depth = "start"
        elif depth == -1:
            depth = "end"
        try:
            name = node[1].__name__
        except BaseException:
            name = str(node[1].__class__).split(".")[-1].split("'")[0]
            if isinstance(node[1], kosh.loaders.core.KoshLoader):
                name = "{}({})".format(name, node[1].feature)
        labels[node] = "{}/{}/{}".format(depth, name, node[0])
    return labels


def draw_execution_graph(G,
                         output_format=None,
                         png_name="kosh_execution_graph.png",
                         clear=True,
                         layout=None):
    """Draws the graph and if provided an output format, draws the shortest path to it
    :param G: networkx graph or KoshExecutionGraph
    :type G: networkx.Graph
    :param output_format: draw shortest path to this format
    :type output_format: str or None
    :param png_name: name of png file to output the graph to
    :type png_name: str
    :param clear: clear matplotlib figure after saving
    :type clear: bool
    :param layout: A dictionary with nodes as keys and positions as values.
                   If not specified a planar layout positioning will be computed.
                   See networkx.drawing.layout for functions that compute node positions.
    :type layout: dict or function
    :returns: None but draws the matplotlib plt is updated and possibly saved
    :rtype: None
    """
    import networkx as nx
    if layout is None:
        layout = nx.planar_layout
    from kosh.exec_graphs import find_network_ends
    if not isinstance(layout, dict):
        layout = layout(G)

    if isinstance(G, kosh.exec_graphs.KoshExecutionGraph):
        G = G.execution_graph()
    lbls_dict = gen_labels(G)
    nx.draw(
        G,
        pos=layout,
        with_labels=True,
        labels=lbls_dict,
        alpha=.5,
        node_size=150,
        edge_color='black',
        style="dashed")
    labels = nx.get_edge_attributes(G, 'weight')
    for k in labels:
        labels[k] = "{:.3g}".format(labels[k])
    nx.draw_networkx_edge_labels(G, layout, edge_labels=labels)
    if output_format is not None:
        starters = find_network_ends(G, start=True, end=False)
        for start in starters:
            pth = nx.shortest_path(
                G, start, (output_format, None, G.seed), weight="weight")
            # build edges
            edges = []
            for i in range(len(pth) - 1):
                edges.append((pth[i], pth[i + 1]))
            nx.draw(
                G,
                pos=layout,
                with_labels=True,
                labels=lbls_dict,
                nodelist=pth,
                edgelist=edges,
                edge_color='red')
    try:
        if "DISPLAY" not in os.environ or os.environ["DISPLAY"] == "":
            import matplotlib
            matplotlib.use("agg", force=True)
        import matplotlib.pyplot as plt
        plt.show()
        plt.savefig(png_name)
        if clear:
            plt.clf()
    except ImportError:
        raise RuntimeError(
            "Could not import matplotlib, will not plot anything")


def compute_fast_sha(uri, n_samples=10):
    """Compute a fast 'almost' unique identifier for a given uri
    Assumes the uri is a path to a file, otherwise simply return hexdigest of md5 on the uri string

    If uri path is valid the 'fast' sha is used by creating an hashlib from
    * file size
    * file first 2kb
    * file last 2kb
    * 2k samples read from `n_samples` evenly spaced in the file

    Warning if size is unchanged and data is changed somewhere else than those samples the sha will be identical
    :param uri: URI to compute fast_sha on
    :type uri: str
    :param n_samples: Number of samples to extract from uri (in addition to beg and end of file)
    :type n_sampe: int
    :return sha: hexdigested sha
    :rtype: str
    """
    import numpy
    if not os.path.exists(uri):
        sha = hashlib.sha256(uri.encode())
        return sha.hexdigest()
    with open(uri, "rb") as f:
        stats = os.fstat(f.fileno())
        size = stats.st_size
        sha = hashlib.sha256("{}".format(size).encode())
        # Create list of start read
        positions = [int(max(x, 0))
                     for x in numpy.linspace(0, size - 2048, n_samples + 2)]
        prev = -1
        for pos in positions:
            # Small file will have multiple times the same bit to read
            if pos != prev:
                # Go there
                f.seek(pos)
                # read some small chunk
                st = f.read(2048)
                prev = pos
            sha.update(st)
    return sha.hexdigest()


def compute_long_sha(uri, buff_size=65536):
    """ Computes sha for a given uri
    :param uri: URI to compute fast_sha on
    :type uri: str
    :param buff_size: How much data to read at once
    :type buff_size: int
    :return sha: hexdigested sha
    :rtype: str
    """
    sha = hashlib.sha256()

    with open(uri, "rb") as f:
        while True:
            st = f.read(buff_size)
            if not st:
                break
                sha.update(st)
    return sha.hexdigest()


def update_store_record(records, store_record):
    """Updates the store record
    :param records: The sina store "records" object
    :type records: sina.datastore.DataStore.RecordOperations
    :param store_record: sina record for store info
    :type store_record: Record
    """
    if hasattr(records, "insert"):
        try:
            records.update(store_record)
        except Exception:  # in case multi-processors interfere with each others
            pass
    return


@lock_strategies.lock_function
def get_store_info_record_attribute(records, attribute):
    """Obtain an attribute from the store record info.
Sometimes when multiple process try to initiate that record it comes back with missing attributes
This will wrap into the lock strategy
    :param records: The sina store "records" object
    :type records: sina.datastore.DataStore.RecordOperations
    :param attribute: The attribute to retrievr
    :type attribute: str
    :returns: sina record for store info
    :rtype: object
    """
    rec = get_store_info_record(records)
    return rec["user_defined"][attribute]


@lock_strategies.lock_function
def get_store_info_record(records):
    """Obtain the sina record containing store info
    If necessary update store to latest standards
    :param records: The sina store "records" object
    :type records: sina.datastore.DataStore.RecordOperations
    :returns: sina record for store info
    :rtype: Record
    """
    from sina.model import Record
    # First let's see if this store contains a dedicated record
    # describing this store specs
    store_info = list(records.find_with_type("__kosh_storeinfo__"))
    if len(store_info) > 1:
        # There is a small chance that the store was created on multiple processors
        # simultaneously and that these are identical, let's try to recover
        # that was true for a small period in Kosh dev branch
        rec = store_info[0]
    elif len(store_info) == 0:
        # ok it's the old type or a new store, let's try to upgrade it for next time
        # and add the store info
        # Because of mpi ranks issues let's fix the id
        rec = Record(id="__kosh_store_info__", type="__kosh_storeinfo__", user_defined={'kosh_information': {}})
        if hasattr(records, "insert"):  # Readonly can't insert
            # It's possible many ranks will try to create this record
            # They are all identical, let's allow the error
            try:
                records.insert(rec)
            except Exception:
                pass
    else:
        rec = store_info[0]
        # This will fail if we get to version x.10
        # revisit then...
        ver = sum(
            [float(x) / 10**i for i, x in enumerate(version().split(".")) if x[0] != 'g'])
        try:
            min_ver = rec["data"]["kosh_min_version"]["value"]
        except Exception:  # Ok rec is somehow missing this info, setting to 0
            min_ver = "0.0"
        min_ver = sum(
            [float(x) / 10**i for i, x in enumerate(min_ver.split("."))])
        if ver < min_ver:
            raise RuntimeError(
                "This Kosh store requires Kosh version greater than {}, you have {}".format(min_ver, version()))
    return rec


def update_store_and_get_info_record(records, ensemble_predicate=None):
    """Obtain the sina record containing store info
    If necessary update store to latest standards
    :param records: The sina store "records" object
    :type records: sina.datastore.DataStore.RecordOperations
    :param ensemble_predicate: The predicate for the relationship to an ensemble
    :type ensemble_predicate: str
    :returns: sina record for store info
    :rtype: Record
    """
    # First let's see if this store contains a dedicated record
    # describing this store specs
    rec = get_store_info_record(records)
    need_update = False
    if "sources_type" not in rec["data"]:
        rec.add_data("sources_type", "file")
        need_update = True
    if "users_type" not in rec["data"]:
        rec.add_data("users_type", "user")
        need_update = True
    if "groups_type" not in rec["data"]:
        rec.add_data("groups_type", "group")
        need_update = True
    if "loaders_type" not in rec["data"]:
        rec.add_data("loaders_type", "koshloader")
        need_update = True
    if "ensembles_type" not in rec["data"]:
        rec.add_data("ensembles_type", "kosh_ensemble")
        need_update = True
    if "ensemble_predicate" not in rec["data"]:
        if ensemble_predicate is None:
            rec.add_data("ensemble_predicate", "is a member of ensemble")
        else:
            rec.add_data("ensemble_predicate", ensemble_predicate)
        need_update = True
    if "cached_features" not in rec["user_defined"]:
        rec["user_defined"]["cached_features"] = collections.OrderedDict()
        need_update = True
    if "kosh_min_version" not in rec["data"]:
        rec.add_data("kosh_min_version", "1.2.1")
        need_update = True
    if "reserved_types" not in rec["data"]:
        rec.add_data("reserved_types", [
            "__kosh_storeinfo__", "file", "user", "group", "kosh_ensemble", "koshloader"])
        need_update = True
    if sorted(rec["data"]["reserved_types"]["value"]) != ['__kosh_storeinfo__', 'file',
                                                          'group', 'kosh_ensemble',
                                                          'koshloader', 'user']:
        rec["data"]["reserved_types"]["value"] = ['__kosh_storeinfo__',
                                                  'file', 'group', 'kosh_ensemble', 'koshloader', 'user']
        need_update = True
    if need_update:
        update_store_record(records, rec)
    return rec


def create_kosh_users(record_handler, users=[os.environ.get("USER", "default"), "anonymous"]):
    """Add Kosh user to the Kosh store
    :param record_handler: The sina records object
    :type record_handler: sina.records
    :param users: list of usernames to add
    :type users: list
    """
    from sina.model import Record
    store_info = list(record_handler.find_with_type(
        ["__kosh_storeinfo__", ]))[0]

    user_type = store_info["data"]["users_type"]["value"]
    # Create users
    for user in users:
        new_user = list(record_handler.find(
            types=[user_type, ], data={"username": user}))
        if len(new_user) == 0:
            uid = hashlib.md5(user.encode()).hexdigest()
            user_record = Record(id=uid, type=user_type, user_defined={'kosh_information': {}})
            user_record.add_data("username", user)
            record_handler.insert(user_record)


def create_new_db(name, db='sql',
                  keyspace=None, **kargs):
    """create_new_db creates a new Kosh database, adds a single user

    :param name: name of database
    :type name: str
    :param db: type of database, defaults to 'sql', can be 'cass'
    :type db: str, optional
    :param keyspace: for cassandra keyspace to use, defaults to None means [user]_k
    :type keyspace: str, optional
    :param kargs: Any additional key/value pairs you need to pass to store creation
    :type kargs: dict
    :return store: An handle to the Kosh store created
    :rtype: KoshStoreClass
    """
    if os.path.exists(name):
        os.remove(name)
    from kosh import connect
    kargs["keyspace"] = keyspace
    kargs["db"] = db
    # Let's remove the now unused arguments
    for key in ["token", "cluster"]:
        if key in kargs:
            warnings.warn(
                "Keyword '{}' is no longer valid, will be ignored".format(key))
            kargs.pop(key)
    store = connect(name, **kargs)
    store.delete_all_contents(force="SKIP PROMPT")
    return store


def version(comparable=False):
    """Returns version string
    :param comparable: returns version as a tuple of ints so it can be compared
    :type comparable: bool
    :returns: version string or tuple
    :rtype: str or tuple
    """
    import pkg_resources
    try:
        __version__ = pkg_resources.get_distribution("kosh").version
    except Exception:
        __version__ = "0.0"
    if comparable:
        tuple_version = ()
        for number in __version__.split("."):
            try:
                tuple_version += (int(number),)
            except ValueError:  # Probably some letter or symbol in here
                pass
        __version__ = tuple_version
    return __version__


def walk_dictionary_keys(dictionary, separator="/"):
    """Walks through a dictionary and return all levels of keys
    sub dictionary keys are append to parent key with the 'separator'
    :param dictionary: The dictionary to walk
    :type dictionary: dict
    :param separator: The string to use between a parent key and its children
    :type separator: str
    :returns: generator of keys and possibly their sub keys
    :rtype: generator
    """
    out = []
    for key in sorted(dictionary.keys(), key=lambda x: str(x)):
        out.append(str(key))
        if isinstance(dictionary[key], dict):
            yld = walk_dictionary_keys(dictionary[key], separator)
            for y in yld:
                st = "{}{}{}".format(key, separator, y)
                out.append(st)
    return out


def get_graph(input_type, loader, transformers):
    """Given a loader and its transformer return path to desired format
    e.g which output format should each transformer pick to be chained to the following one
    in order to obtain the desired outcome for format
    :param input_type: input type of first node
    :type input_type: str
    :param loader: original loader
    :type loader: KoshLoader
    :param transformers: set of transformers to be added after loader exits
    :type transformers: list of KoshTransformer
    :returns: execution graph
    :rtype: networkx.OrderDiGraph
    """
    import networkx as nx
    from kosh.exec_graphs import populate
    if input_type not in loader.types:
        raise RuntimeError(
            "loader cannot load mime_type {}".format(input_type))
    try:
        G = nx.OrderedDiGraph()
    except AttributeError:  # networkx 3.0 removed OrderedDiGraph
        G = nx.DiGraph()
    G.seed = random.random()
    start_node = (input_type, loader, G.seed)  # so each graph is unique
    G.add_node(start_node)
    if len(transformers) == 0:
        # No transformer
        for out_format in loader.types[input_type]:
            node = (out_format, None, G.seed)
            G.add_edge(start_node, node)
    else:
        populate(
            G,
            start_node,
            loader.types[input_type],
            transformers)
    return G


def cleanup_sina_record_from_kosh_sync(record):
    """Kosh adds data in the 'user_defined' section of records to keep track of syncing
    This removes these attributes
    :param record: The Sina record to cleanup
    :type record: sina.model.Record
    :return: json loaded representation of the record
    :rtype: dict"""
    # cleanup the record
    record["user_defined"]['kosh_information'].pop("last_update_from_db", None)
    for key in list(record["user_defined"]['kosh_information'].keys()):
        if key.endswith("last_modified"):
            record["user_defined"]['kosh_information'].pop(key)
    return orjson.loads(record.to_json())


def record_to_dataset(record):
    """Converts a Sina record to a KoshDatset
    :param record: The Sina Record to convert
    :type record: sina.model.Record
    :return: kosh version of the record
    :rtype: kosh.KoshDatset"""
    temp_store = kosh.connect(None)
    temp_store.get_sina_records().insert(record)
    return next(temp_store.find())


def record_to_dataframe(rec_uri):
    """Converts a Sina record to a Pandas DataFrame
    :param rec_uri: The Sina Record URI to convert
    :type rec_uri: str
    :return: Pandas DataFrame version of the record
    :rtype: Pandas.DataFrame"""
    temp_store = kosh.connect(None)
    temp_store.import_dataset(rec_uri, match_attributes=['id'])
    return temp_store.to_dataframe()


def datasets_in_place_of_records(func):
    """This decorator will convert all Record input or output to KoshDataset
    This allows a user to use sina functions that expect Record with Kosh datasets instead"""

    from sina.model import Record

    def wrapper(*args, **kwargs):
        new_args = [x.get_record() if isinstance(x, kosh.KoshDataset) else x for x in args]
        new_kwargs = {}
        for k, v in kwargs.items():
            new_kwargs[k] = v.get_record() if isinstance(v, kosh.KoshDataset) else v
        out = func(*new_args, **new_kwargs)
        if hasattr(out, "__iter__"):
            return [record_to_dataset(v) if isinstance(v, Record) else v for v in out]
        else:
            return record_to_dataset(out) if isinstance(out, Record) else out
    return wrapper


def update_json_file_with_records_and_relationships(file, output_dict):
    if file is not None:
        if os.path.exists(file):
            with open(file) as f:
                file_dict = orjson.loads(f.read())
            records_already_in_file = file_dict["records"]
            # You can't "set" records
            # This erased records
            file_dict.update(output_dict)
            # ids that are now in file_dict
            new_dataset_rec_ids = [x["id"] for x in output_dict["records"]]
            # Make sure we put the records that were in the file back in
            for record in records_already_in_file:
                if record["id"] not in new_dataset_rec_ids:
                    file_dict["records"].append(record)
        else:
            file_dict = output_dict

        with open(file, "w") as f:
            rels = []
            for relationship in file_dict['relationships']:
                rels.append(relationship.__dict__)
            file_dict['relationships'] = rels
            try:
                f.write(orjson.dumps(file_dict).decode())
            except AttributeError:
                f.write(orjson.dumps(file_dict))
    return output_dict


def __check_valid_connection_type__(connection_type, connection_types_allowed):
    """Checks if the `connection_type` is in `connection_types_allowed` for different stores

    :param connection_types_allowed: The connection types allowed for this store type
    :type connection_types_allowed: lst
    """
    if connection_type not in connection_types_allowed:
        raise RuntimeError(f"This functionality is allowed for {connection_types_allowed} only stores.")



def auto_valid(validation):
    """auto_valid converts class type
    to their 'isinstance' function in the future other
    keyword maybe added to the list

    :param validation: class def or type
    :type validation: type
    :return: isinstance lambda function or validation function
    :rtype: function
    """
    if isinstance(validation, type):
        def lmb(x): return isinstance(x, validation)
        return lmb
    return validation


def validate_value(value, validation):
    """validate_value Validates a value for a given validation function (or list of)

    :param value: value to validate
    :type value: any
    :param validation: validation function (or list of)
    :type validation: callable
    :raises ValueError: value does not vaidate
    :return: True if validates
    :rtype: bool
    """
    if validation is None:
        return True
    validation = auto_valid(validation)
    if callable(validation):
        res = validation(value)
        if not res:
            raise ValueError("value {value} failed validation".format(value=value))
    elif isinstance(validation, list):
        success = False
        for item in validation:  # Loop thru elements to get a possible success
            item = auto_valid(item)
            if callable(item):
                try:
                    res = item(value)
                    success += res
                    if success:
                        break  # no need to check further
                except Exception:
                    pass
            elif value == item:
                success += True
                break  # no need to check further

        if not success:  # All options failed
            raise ValueError("Could not validate value '{value}'".format(value=value))
    return True


class KoshSchema(object):
    def __init__(self, required={}, optional={}):
        """Schema for Kosh objects
        :param required: Dictionary of required keys and their validation function(s).
        :type required: dict
        :param optional: Dictionary of optional keys and their validation function(s)
        :type optional: dict
        Validation functions must be pickable.
        If multiple validation functions are provided (in a list),
        then the attribute is valid if ANY function return True.
        Validation function must return True, False or raise an Exception.
        """
        self.required = required
        self.optional = optional

    def __str__(self):
        """string representation"""
        st = """Kosh Validation Object
        Required attributes and their validations:
        {}
        Optional attributes and their validations:
        {}""".format(self.required, self.optional)
        return st

    def validate(self, obj, level=None):
        """validate an object through a schema
        Checks that the obj has all the required attribute
        and that both required and present optional attributes pass
        validation functions.

        :param obj: object to validate
        :type obj: str
        :param level: Dataset or Ensemble level, used for `ensemble_tags`
        :type level: str
        :raises ValueError: obj does not validate through the schema
        :return: True if validates
        "rtype: bool
        """
        # First check all the required keys
        req_errors = {}
        for k, v in self.required.items():
            try:
                if isinstance(obj, dict):
                    value = obj[k]
                else:
                    value = getattr(obj, k)
                validate_value(value, v)
            except Exception as err:
                req_errors[k] = err
        opt_errors = {}
        for k, v in self.optional.items():
            try:
                if isinstance(obj, dict):
                    value = obj[k]
                else:
                    value = getattr(obj, k)
            except AttributeError:
                continue
            try:
                validate_value(value, v)
            except Exception as err:
                opt_errors[k] = err

        if len(req_errors) != 0 or len(opt_errors) != 0:
            if isinstance(obj, dict):
                obj_level = level
            else:
                obj_level = obj.id
            raise ValueError(
                "Could not validate {}\n"
                "{} required attribute errors: {}\n"
                "{} optional attributes errors: {}".format(
                    obj_level, len(req_errors), req_errors, len(opt_errors), opt_errors))
        return True

    def validate_attribute(self, attribute, value):
        """validate_attribute validates a value for a specific attribute

        :param attribute: attribute to validate
        :type attribute: str
        :param value: value to validate
        :type value: any
        :return: True or False
        "rtype: bool
        """
        if attribute in self.required:
            result = validate_value(value, self.required[attribute])
        elif attribute in self.optional:
            result = validate_value(value, self.optional[attribute])
        else:
            result = True
        return result



from .core import KoshOperator


class KoshLNorm(KoshOperator):
    """Interpolates Two 2-D Arrays [x0, y0], [xy, y1] or four 1-D arrays x0, y0, x1, y1
    to a shared domain and calculates the `LNorm = np.sum(abs(y0-y1)**power)**1/power`.
    Returns the shared x-axis, the interpolated y-axis for both features, and the LNorm.
    """
    types = {"numpy": ["numpy", ],
             "dict": ["numpy", ],
             "hdf5": ["numpy", ]}

    def __init__(self, *args, **options):
        """
        :param inputs: Two 2-D Arrays [x0, y0], [xy, y1] or four 1-D arrays x0, y0, x1, y1 features or data
        :type inputs: kosh feature or data
        :param power: The power of the `LNorm = np.sum(abs(y0-y1)**power)**1/power`
        :type power: int
        :param overlap_only: Only interpolate within the overlap of both x-axis.
                             Eliminates the need for `left`, `right`, and `period` parameters below.
        :param overlap_only: bool
        :param left: The `left` parameter in `numpy.interp()`.
                     "Value to return for x < xp[0], default is fp[0]."
        :type left: float or complex
        :param right: The `right` parameter in `numpy.interp()`.
                      "Value to return for x > xp[-1], default is fp[-1]."
        :type right: float or complex
        :param period: The `period` parameter in `numpy.interp()`.
                       "A period for the x-coordinates. This parameter allows the proper interpolation of angular
                       x-coordinates. Parameters left and right are ignored if period is specified."
        :type period: None or float
        :returns:
            - x (:py:class:`ndaray`) - The shared domain x-axis
            - y0 (:py:class:`ndarray`) - The first shared domain interpolated y-axis
            - y1 (:py:class:`ndarray`) - The second shared domain interpolated y-axis
            - LNorm (:py:class:`float`) - The LNorm of the overlapping data points
        """
        super(KoshLNorm, self).__init__(*args, **options)
        self.options = options
        loaders_used = {}
        for i, arg in enumerate(args):
            loaders_used[i] = {}
            loaders_used[i]['start_nodes'] = arg.__dict__['start_nodes']
            loaders_used[i]['end_nodes'] = arg.__dict__['end_nodes']
        self.loaders_used = loaders_used

    def operate(self, *inputs, **kargs):
        import numpy as np

        features = {}
        features[0] = {}
        features[1] = {}

        # Gather data from inputs
        if len(inputs) == 2:

            for i in range(2):

                # UltraLoader dict
                if (self.loaders_used[i]['start_nodes'][0][0] == 'ultra' and
                        self.loaders_used[i]['end_nodes'][0][0] == 'dict'):

                    for key, value in inputs[i].items():
                        features[i]['x-axis'] = value['x-axis']
                        features[i]['y-axis'] = value['y-axis']

                # HDF5Loader HDF5
                elif (self.loaders_used[i]['start_nodes'][0][0] == 'hdf5' and
                      self.loaders_used[i]['end_nodes'][0][0] == 'numpy'):
                    features[i]['x-axis'] = inputs[i][0]
                    features[i]['y-axis'] = inputs[i][1]

                # 2D Numpy
                elif (self.loaders_used[i]['end_nodes'][0][0] == 'numpy'):
                    features[i]['x-axis'] = inputs[i][0]
                    features[i]['y-axis'] = inputs[i][1]

        elif len(inputs) == 4:

            # HDF5Loader HDF5 and 2D Numpy
            features[0]['x-axis'] = inputs[0]
            features[0]['y-axis'] = inputs[1]
            features[1]['x-axis'] = inputs[2]
            features[1]['y-axis'] = inputs[3]

        else:
            raise ValueError("Inputs must be two 2-D Arrays [x0, y0], [xy, y1] or four 1-D arrays x0, y0, x1, y1")

        # Acquire common domain
        if self.options.get("overlap_only"):
            x = list(features[0]['x-axis'][np.where(
                np.logical_and(features[0]['x-axis'] >= features[1]['x-axis'][0],
                               features[0]['x-axis'] <= features[1]['x-axis'][-1]))])
            x.extend(list(features[1]['x-axis'][np.where(
                np.logical_and(features[1]['x-axis'] >= features[0]['x-axis'][0],
                               features[1]['x-axis'] <= features[0]['x-axis'][-1]))]))
            x = list(set(x))
            x.sort()
        else:
            x = []
            x.extend(features[0]['x-axis'])
            x.extend(features[1]['x-axis'])
            x = list(set(x))
            x.sort()

        left = self.options.get("left")
        right = self.options.get("right")
        period = self.options.get("period")
        y0 = np.interp(x, features[0]['x-axis'], features[0]['y-axis'],
                       left=left, right=right, period=period)
        y1 = np.interp(x, features[1]['x-axis'], features[1]['y-axis'],
                       left=left, right=right, period=period)

        power = self.options.get("power", 2)
        LNorm = np.sum(abs(y0-y1)**power)**1/power
        return np.array(x), np.array(y0), np.array(y1), LNorm



from kosh.transformers import kosh_cache_dir
import os
import hashlib
import pickle
import six
from kosh.exec_graphs import KoshExecutionGraph
from ..core_sina import KoshSinaObject, KoshSinaFile
from ..utils import get_graph, find_curveset_and_curve_name
from ..dataset import KoshDataset
from ..ensemble import KoshEnsemble


class KoshGenericObjectFromFile(KoshSinaObject):
    """Kosh object pointing to a file"""

    def __init__(self, uri, mode, obj, *args, **kwds):
        super(KoshGenericObjectFromFile, self).__init__(obj.id, obj.__store__,
                                                        obj.__type__, obj.__record_handler__, protected=["file_obj"])
        self.file_obj = open(uri, mode, *args, **kwds)

    def __enter__(self):
        return self.file_obj

    def __exit__(self, *args):
        self.file_obj.close()

    def get(self, *args, **kargs):
        """Reads the file all arguments are ignored"""
        return self.file_obj.read()


class KoshLoader(KoshExecutionGraph):
    """
    :param types: types is a dictionary on known type that can be loaded
    as key and export format as value, defaults to {"dataset": []}
    :type types: dict
    """
    types = {"dataset": []}

    def __init__(self, obj, mime_type=None, uri=None, requestorId=None):
        """KoshLoader generic Kosh loader
        :param obj: object the loader will try to load from
        :type obj: object
        :param mime_type: If you want to force the mime_type to use
        :type mime_type: str
        :param uri: If you want/need to force the uri to use
        :type uri: str
        :param requestorId: The id of the dataset requesting data
        :type requestorId: str
        """
        self.requestorId = requestorId
        self.signature = hashlib.sha256(repr(self.__class__).encode()).hexdigest()
        self.signature = self.update_signature(obj.id)
        if mime_type is None:
            self._mime_type = obj.mime_type
        else:
            self._mime_type = mime_type
        if uri is None:
            try:
                self.uri = obj.uri
            except AttributeError:  # Not uri on this
                self.uri = None
        else:
            self.uri = uri
        rec = obj.__store__.get_record(obj.id)
        if (rec["type"] not in obj.__store__._kosh_reserved_record_types and mime_type is None)\
                or rec["type"] in ["__kosh_storeinfo__", obj.__store__._ensembles_type]:
            self._mime_type = "dataset"
        if self._mime_type not in self.types:
            open_anything = False
            for t in self.types:
                if t == "dataset":  # special skipping
                    continue
                if len(self.types[t]) == 0:
                    open_anything = True
            if not open_anything:
                raise RuntimeError(
                    "will not be able to load object of type {mime_type}".format(mime_type=self._mime_type))
        self.obj = obj
        self.__records__ = obj.__store__.get_sina_records()

    def get_requestor(self):
        """Returns the Kosh object requesting the data"""
        try:
            return self.obj.__store__.open(self.requestorId)
        except Exception:  # some kosh object do not have an open function
            return self.obj.__store__._load(self.requestorId)

    def known_types(self):
        """Lists types of Kosh objects this loader can handle

        :return: list of Kosh type this loader can handle
        :rtype: list
        """
        return list(self.types.keys())

    def known_load_formats(self, atype):
        """Lists all the formats this loader knows how to export to for a given type

        :param atype: type we wish to to the formats for
        :type format: str
        :return: list of format this type can be exported to by the loader
        :rtype: list
        """
        return self.types.get(atype, [])

    def open(self, mode="r"):
        """Open function
        :param mode: mode to open the object in
        :type mode: str
        :return: opened object
        :rtype: object"""
        return self

    def update_signature(self, *args, **kargs):
        """Updated the signature based to a set of args and kargs
        :param *args: as many arguments as you want
        :type *args: list
        :param **kargs: key=value style arguments
        :type **kargs: dict
        :return: updated signature
        :rtype: str
        """
        signature = hashlib.sha256(self.signature.encode())
        for arg in args:
            signature.update(repr(arg).encode())
        for kw in kargs:
            signature.update(repr(kw).encode())
            signature.update(repr(kargs[kw]).encode())
        return signature.hexdigest()

    def get_execution_graph(self, feature, transformers=[]):
        """Generates the execution graph to extract a feature and possibly transform it.

        :param feature: desired feature
        :type feature: str
        :param format: desired output format
        :type format: str
        :param transformers: A list of transformers to use after the data is loaded
        :type transformers: kosh.transformer.KoshTranformer
        :return: execution graph to get to the possibly transformed feature
        :rtype: networkx.OrderDiGraph
        """
        # Let's get the execution path
        G = get_graph(self._mime_type, self, transformers)
        return G

    def get(self, feature, format=None, transformers=[],
            use_cache=True, cache_file_only=False, cache_dir=None,
            **kargs):
        """Extracts a feature and possibly transforms it
        :param feature: desired feature
        :type feature: str
        :param format: desired output format
        :type format: str
        :param transformers: A list of transformers to use after the data is loaded
        :type transformers: kosh.transformer.KoshTranformer
        :param use_cache: Try to use cached data if available
        :type use_cache: bool
        :param cache_file_only: If True, simply return name of cache_file
        :type cache_file_only: bool
        :param cache_dir: where do we cache the result?
        :type cache_dir: str
        **kargs will be stored on loader object
        format and feature are stored on the object for extraction by extraction functions
        This function calls first the loader's preprocess function
        This is followed by an actual data extraction via the 'extract' function
        Finally 'postprocess' is called on the extracted data

        Reserved keyword:
        preprocess: function use to preprocess (default to self.preprocess)
        postprocess: function use to postprocess (default to self.postprocess)
        """

        if cache_dir is None:
            cache_dir = kosh_cache_dir
        self.cache_dir = cache_dir
        self.use_cache = use_cache
        self.cache_file_only = cache_file_only
        self.feature = feature
        self._user_passed_parameters = (None, kargs)
        G = self.get_execution_graph(feature, transformers=transformers)
        return KoshExecutionGraph(G).traverse(format=format, **kargs)

    def extract_(self, format):
        if format is None:
            format = self.types[self._mime_type][0]
        if len(self.types) != 0 and format not in self.types[self._mime_type]:
            raise ValueError("Loader cannot output type {self._mime_type} to {format} format".format(
                self=self, format=format))
        self.format = format
        _, kargs = self._user_passed_parameters
        signature = self.update_signature(self.feature, format, **kargs)
        if self.cache_file_only is True:
            # Ok user just wants to know where cache should be (plus/minus extesions)
            return signature
        # Let's generate the signatures for each step of the path.
        # And try to load it
        cache_success = False
        if self.use_cache:
            try:
                data = self.load(signature)
                cache_success = True
            except Exception:
                pass
            if cache_success:
                return data

        kargs.get("preprocess", self.preprocess)()
        data = self.extract()
        data = kargs.get("postprocess", self.postprocess)(data)
        return data

    def save(self, cache_file, content):
        """Pickles some data to a cache file
        :param cache_file: name of cache file, will be joined with self.cache_dir
        :type cache_file: str
        :param content: content to save to cache
        :type content: object
        """
        with open(os.path.join(self.cache_dir, cache_file), "wb") as f:
            pickle.dump(content, f)

    def load(self, cache_file):
        """Loads content from cache
        :param cache_file: name of cache file, will be joined with self.cache_dir
        :type cache_file: str
        :return: unpickled data
        :rtpye: object
        """
        with open(os.path.join(self.cache_dir, cache_file), "rb") as f:
            data = pickle.load(f)
        return data

    def _list_features(self, *args, **kargs):
        """Wrapper on top of list_features to snatch from cache rather than calling every time"""
        verbose = kargs.get("verbose", False)
        use_cache = kargs.pop("use_cache", True)
        __listed_features_cache = self.obj.__store__._cached_features_
        __listed_features = None
        if __listed_features_cache is not None and use_cache:
            if self.uri is None or self.uri == "":
                __listed_features = __listed_features_cache.get(self.signature+self.obj.id, None)
            else:
                __listed_features = __listed_features_cache.get(self.signature+self.uri, None)
        if __listed_features is None:  # didn't get it from cache
            try:
                __listed_features = self.list_features(*args, **kargs)
                # Reset
                if self.uri is None or self.uri == "":
                    __listed_features_cache[self.signature+self.obj.id] = __listed_features
                else:
                    __listed_features_cache[self.signature+self.uri] = __listed_features
                # Update the local stroe features cache
                self.obj.__store__._cached_features_ = __listed_features_cache
            except Exception as err:
                if verbose:
                    print("\tCould not obtain features from loader {}\n\t\tError:{}".format(self, err))
                # Broken loader at the moment
                __listed_features = []
        out = __listed_features
        return out

    def list_features(self, *args, **kargs):
        """list_features Given the obj it's loading return a list of features (variables)
        it can extract

        :return: list of available features from this loader
        :rtype: list
        """
        return []

    def describe_feature(self, feature):
        """describe_feature describe the feature as a dictionary

        :param feature: feature to describe
        :type feature: str
        :return: dictionary with attributes describing the feature
        :rtype: dict
        """
        raise NotImplementedError("describe_feature method not implemented")

    def preprocess(self):
        """preprocess sets things up for the extract function

        This should be preceeded by a call to 'get' which stored its args
        in self._user_passed_parameters
        """
        return

    def extract(self):
        """extract this function does the heavy lifting of the extraction
        it needs to be implemented by each loader.

        We recommend returning pointer to the data as much as possible

        :raises NotImplementedError:
        """
        raise NotImplementedError

    def postprocess(self, data):
        """postprocess Given the extracted data apply some post processing to it

        :param data: result of the extract function
        :type data: any
        :return: post processed
        :rtype: any
        """
        return data


class KoshFileLoader(KoshLoader):
    """Kosh loader to load content from files"""
    types = {"file": []}

    def __init__(self, obj, **kwargs):
        super(KoshFileLoader, self).__init__(obj, **kwargs)

    def open(self, mode='r'):
        """open/load the matching Kosh Sina File

        :param mode: mode to open the file in, defaults to 'r'
        :type mode: str, optional
        :return: Kosh File object
        """
        return KoshGenericObjectFromFile(self.uri, mode, self.obj)

    def extract(self, feature, format):
        """extract return a feature from the loaded object.

        :param feature: variable to read from file
        :type feature: str
        :param format: desired output format
        :type format: str
        :return: data
        """
        with open(self.uri) as f:
            return f.read()

    def list_features(self, *args, **kargs):
        """list_features list features in file,

        :return: list of features available in file
        :rtype: list
        """
        return []

    def describe_feature(self, feature):
        """describe a feature

        :param feature: feature (variable) to read, defaults to None
        :type feature: str, optional if loader does not require this
        :return: dictionary describing the feature
        :rtype: dict
        """
        if feature not in self.list_features():
            raise ValueError("feature {feature} is not available".format(feature=feature))
        return {}


class KoshSinaLoader(KoshLoader):
    """Sina base class for loaders"""
    types = {"dataset": ["numpy", ]}

    def __init__(self, obj, **kargs):
        """KoshSinaLoader generic sina-based loader
        """
        super(KoshSinaLoader, self).__init__(obj, **kargs)

    def open(self, *args, **kargs):
        """open the object
        """
        record = self.obj.__store__.get_record(self.obj.id)
        if record["type"] not in self.obj.__store__._kosh_reserved_record_types:
            return KoshDataset(
                self.obj.id, store=self.obj.__store__, record=record)
        if record["type"] == self.obj.__store__._sources_type:
            return KoshSinaFile(
                self.obj.id, store=self.obj.__store__, record=record)
        elif record["type"] == self.obj.__store__._ensembles_type:
            return KoshEnsemble(
                self.obj.id, store=self.obj.__store__, record=record)
        else:
            return KoshSinaObject(self.obj.id, self.obj.__store__, record["type"], protected=[
            ], record_handler=self.obj.__store__.__record_handler__, record=record)

    def list_features(self):
        record = self.obj.__store__.get_record(self.obj.id)
        # Using set in case a variable is both in independent and dependent
        # Dependent would win when getting the data
        curves = ()
        for curve in record["curve_sets"]:
            curves += ((curve, None), )
            for curve_type in ["independent", "dependent"]:
                for name in record["curve_sets"][curve][curve_type]:
                    curves += ((curve, name), )
        joined = sorted([x if y is None else x+"/"+y for x, y in curves])
        if joined == sorted(set(joined)):
            return joined
        else:
            return sorted(curves, key=lambda x: (x[0], "" if x[1] is None else x[1]))

    def extract(self, *args, **kargs):
        import numpy
        features = self.feature
        if not isinstance(features, list):
            features = [self.feature, ]
        record = self.obj.__store__.get_record(self.obj.id)
        out = []
        for feature in features:
            if isinstance(feature, six.string_types):
                possibilities = find_curveset_and_curve_name(feature, record)
                if len(possibilities) > 1:
                    raise ValueError("Could not uniquely resolve {} if to could belong to any of: {}".format(
                        feature, possibilities))
                curve_set, curve_name = possibilities[0]
            else:
                curve_set, curve_name = feature
            # Here we are assuming the curve root name cannot have "/" in it
            curve_set = record["curve_sets"][curve_set]
            if curve_name is not None:
                if curve_name in curve_set["independent"]:
                    curve = curve_set["independent"][curve_name]["value"]
                elif curve_name in curve_set["dependent"]:
                    curve = curve_set["dependent"][curve_name]["value"]
                else:
                    raise ValueError("Cannot find curve {} in curve_set {}".format(curve_name, curve_set))
                out.append(numpy.array(curve))
            else:
                # we want all curves
                all = []
                # Matching order (indep/dep) that we used in list_features
                for curve_type in ["independent", "dependent"]:
                    # Same order as list_features()
                    for curve_name in sorted(curve_set[curve_type].keys()):
                        curve = curve_set[curve_type][curve_name]["value"]
                        all.append(numpy.array(curve))
                out.append(all)
        if not isinstance(self.feature, list):
            return out[0]
        else:
            return out
